import COM.hugin.HAPI.*;
import java.awt.Point;
import java.io.*;

/**
 * Class for setting up and running a BOBW network, with or without additional content features
 * 
 * @author Cecilie Be
 *
 */
public class BOBW extends Domain implements ParseListener {

	private LabelledDCNode user;
	private LabelledDCNode zip;
	private LabelledDCNode age;
	private LabelledDCNode movie;
	private LabelledDCNode year;
	private LabelledDCNode director;
	private LabelledDCNode actor;
	private LabelledDCNode community;
	private LabelledDCNode category;
	private ContinuousChanceNode rating;

	private int numberOfComCatStates;
	private boolean normalized;
	private int numberOfTrainingCases;
	private int numberOfTestCases;
	private int numberOfUsers;
	private int numberOfMovies;
	private String modelType;//plain,year,dir&act,zip,age
	private String trainingSet;
	private String testSet;
	private int[] iterations;

	private int numberOfYearStates;
	private int numberOfDirectorStates;
	private int numberOfActorStates;
	private int numberOfZipStates;
	private int numberOfAgeStates;

	private double[] results;
	private double[] averageTable;

	/**
	*
	* Constructs a new network with a user, community, movie, category and rating node, where the four first are LabelledDCNodes
	* and the last is a ContinuousChanceNode
	* 
	* Sets the ratings to unnormalized by default. This can be changed to normalized by calling the setNormalized()-method 
	* 
	* Sets the number of EM iterations to one round of ten non-continuous 
	* learning iterations and one one continuous learning iteration by default. This can be changed to the desired
	* combination by calling the setNumberOfIterations()-method
	*
	* @param modelType The model type is "plain" if no additional features are to be included in the network, "year" to include
	* movie release year, "dir&act" to include directors and actors, "zip" to include user zip codes, and "age" to include user age
	* @param trainingSet The path to the training set file
	* @param numberOfTrainingCases The number of data cases in the training set
	* @param testSet The path to the test set file
	* @param numberOfTestCases The number of data cases in the test set
	* @param comCatStates The number of states which is wanted in the two latent nodes Community and Category
	*
	*/
	public BOBW(String modelType, String trainingSet, int numberOfTrainingCases, String testSet, int numberOfTestCases, int numberOfUsers, int numberOfMovies, int comCatStates) throws ExceptionHugin {
		super();
		iterations = new int[3];
		setNumberOfIterations(10,1,1);//default
		setNormalized(false);//default
		setModelType(modelType);
		setTrainingSet(trainingSet);
		setNumberOfTrainingCases(numberOfTrainingCases);
		setTestSet(testSet);
		setNumberOfTestCases(numberOfTestCases);
		setNumberOfUsers(numberOfUsers);
		setNumberOfMovies(numberOfMovies);
		setNumberOfComCatStates(comCatStates);
		if (modelType.equals("year")) {
			setNumberOfYearStates(4);//default
		}
		else if (modelType.equals("dir&act")) {
			setNumberOfDirectorStates(389);//default
			setNumberOfActorStates(242);//default
		}
		else if (modelType.equals("zip")) {
			setNumberOfZipStates(10);//default
		}
		else if (modelType.equals("age")) {
			setNumberOfAgeStates(4);//default
		}
	}

	private void setModelType(String type) {
		if (!((type.equals("plain"))||(type.equals("year"))||(type.equals("dir&act"))||(type.equals("zip"))||(type.equals("age")))) {
			System.out.println("You have stated an illegal model type!");
			System.out.println("Please close the program and try again.");
		}
		else {
			modelType = type;
		}
	}
	private void setNumberOfYearStates(int number) {
		numberOfYearStates = number;
	}
	private void setNumberOfDirectorStates(int number) {
		numberOfDirectorStates = number;
	}
	private void setNumberOfActorStates(int number) {
		numberOfActorStates = number;
	}
	private void setNumberOfZipStates(int number) {
		numberOfZipStates = number;
	}
	private void setNumberOfAgeStates(int number) {
		numberOfAgeStates = number;
	}
	private void setTrainingSet(String set) {
		trainingSet = set;
	}
	private void setNumberOfTrainingCases(int number) {
		numberOfTrainingCases = number;
	}
	private void setTestSet(String set) {
		testSet = set;
	}
	private void setNumberOfTestCases(int number) {
		numberOfTestCases = number;
	}
	private void setNumberOfUsers(int number) {
		numberOfUsers = number;
	}
	private void setNumberOfMovies(int number) {
		numberOfMovies = number;
	}
	private void setNumberOfComCatStates(int number) {
		numberOfComCatStates = number;
	}

	/**
	*
	* Sets the number of EM iterations to be used for the learning of the model (if other than the default is wanted)
	*
	* @param upperNetwork The number of EM iterations on the non-continuous nodes for each round
	* @param ratingNode The number of EM iterations on the continuous node for each round
	* @param rounds the number of times the iterations on the non-continuous and continuous nodes are to be repeated
	*
	*/
	public void setNumberOfIterations(int upperNetwork, int ratingNode, int rounds) {
		iterations[0] = upperNetwork;
		iterations[1] = ratingNode;
		iterations[2] = rounds;
	}

	/**
	*
	* Sets the ratings to normalized or unnormalized (they are unnormalized per default)
	*
	* @param norm true if you want normalized ratings, false otherwise
	*
	*/
	public void setNormalized(boolean norm) {
			normalized = norm;
			if (normalized) {
				averageTable = new double[numberOfUsers];
				computeUserAverage();
			}
	}

	/**
	* The method called to set up the network with the chosen parameters, train it with the given training set,
	* and test it with respect to the given test set
	*
	* @return returns a double-array where the first entry is the MAE, and the second is the RMSE
	*
	*/
	public double[] runExternal() {
		results = new double[2];
		setUpNetwork();
		runTraining();
		return results;
	}

	/**
	*
	* Initializes the necessary nodes with the correct number of states
	*
	*/
	private void setUpNetwork() {
		try {

			double[] testDouble;
			double testDoubleElement;

			user = new LabelledDCNode(this);
			user.setName("user");
			user.setLabel("User");
			user.setPosition(new Point(50,250));
			user.setNumberOfStates(numberOfUsers);
			for (int i=0; i<user.getNumberOfStates(); i++) {
				user.setStateLabel(i, ""+(i+1));
			}
			testDouble = new double[user.getNumberOfStates()];
			testDoubleElement = 0;
			for (int i=0; i<user.getNumberOfStates(); i++) {
				testDoubleElement = Math.random();
				testDouble[i] = testDoubleElement;
			}
			user.getTable().setData(testDouble);

			movie = new LabelledDCNode(this);
			movie.setName("movie");
			movie.setLabel("Movie");
			movie.setPosition(new Point(250,250));
			movie.setNumberOfStates(numberOfMovies);
			for (int i=0; i<movie.getNumberOfStates(); i++) {
				movie.setStateLabel(i, ""+(i+1));
			}
			testDouble = new double[movie.getNumberOfStates()];
			testDoubleElement = 0;
			for (int i=0; i<movie.getNumberOfStates(); i++) {
				testDoubleElement = Math.random();
				testDouble[i] = testDoubleElement;
			}
			movie.getTable().setData(testDouble);

			community = new LabelledDCNode(this);
			community.setName("community");
			community.setLabel("Community");
			community.setPosition(new Point(50,125));
			community.setNumberOfStates(numberOfComCatStates);
			for (int i=0; i<numberOfComCatStates; i++) {
				community.setStateLabel(i, ""+i);
			}
			community.addParent(user);
			testDouble = new double[numberOfComCatStates*user.getNumberOfStates()];
			testDoubleElement = 0;
			for (int i=0; i<testDouble.length; i++) {
				testDoubleElement = Math.random();
				testDouble[i] = testDoubleElement;
			}
			community.getTable().setData(testDouble);

			if (modelType.equals("zip")) {
				zip = new LabelledDCNode(this);
				zip.setName("zip");
				zip.setLabel("Zip");
				zip.setPosition(new Point(10,125));
				zip.setNumberOfStates(numberOfZipStates);
				for (int i=0; i<zip.getNumberOfStates(); i++) {
					zip.setStateLabel(i, ""+i);
				}
				zip.addParent(community);
				testDouble = new double[zip.getNumberOfStates()*community.getNumberOfStates()];
				testDoubleElement = 0;
				for (int i=0; i<(zip.getNumberOfStates()*community.getNumberOfStates()); i++) {
					testDoubleElement = Math.random();
					testDouble[i] = testDoubleElement;
				}
				zip.getTable().setData(testDouble);
			}

			if (modelType.equals("age")) {
				age = new LabelledDCNode(this);
				age.setName("age");
				age.setLabel("Age");
				age.setPosition(new Point(10,125));
				age.setNumberOfStates(numberOfAgeStates);
				for (int i=0; i<age.getNumberOfStates(); i++) {
					age.setStateLabel(i, ""+(i+1));
				}
				age.addParent(community);
				testDouble = new double[age.getNumberOfStates()*community.getNumberOfStates()];
				testDoubleElement = 0;
				for (int i=0; i<(age.getNumberOfStates()*community.getNumberOfStates()); i++) {
					testDoubleElement = Math.random();
					testDouble[i] = testDoubleElement;
				}
				age.getTable().setData(testDouble);
			}

			category = new LabelledDCNode(this);
			category.setName("category");
			category.setLabel("Category");
			category.setPosition(new Point(250,125));
			category.setNumberOfStates(numberOfComCatStates);
			for (int i=0; i<category.getNumberOfStates(); i++) {
				category.setStateLabel(i, ""+i);
			}
			category.addParent(movie);
			testDouble = new double[category.getNumberOfStates()*movie.getNumberOfStates()];
			testDoubleElement = 0;
			for (int i=0; i<testDouble.length; i++) {
				testDoubleElement = Math.random();
				testDouble[i] = testDoubleElement;
			}
			category.getTable().setData(testDouble);

			if (modelType.equals("year")) {
				year = new LabelledDCNode(this);
				year.setName("year");
				year.setLabel("Year");
				year.setPosition(new Point(10,125));
				year.setNumberOfStates(numberOfYearStates);
				for (int i=0; i<year.getNumberOfStates(); i++) {
					year.setStateLabel(i, ""+(i+1));
				}
				year.addParent(category);
				testDouble = new double[year.getNumberOfStates()*category.getNumberOfStates()];
				testDoubleElement = 0;
				for (int i=0; i<(year.getNumberOfStates()*category.getNumberOfStates()); i++) {
					testDoubleElement = Math.random();
					testDouble[i] = testDoubleElement;
				}
				year.getTable().setData(testDouble);
			}

			if (modelType.equals("dir&act")) {
				director = new LabelledDCNode(this);
				director.setName("director");
				director.setLabel("Director");
				director.setPosition(new Point(10,125));
				director.setNumberOfStates(numberOfDirectorStates);
				for (int i=0; i<director.getNumberOfStates(); i++) {
					director.setStateLabel(i, ""+(i+1));
				}
				director.addParent(category);
				testDouble = new double[director.getNumberOfStates()*category.getNumberOfStates()];
				testDoubleElement = 0;
				for (int i=0; i<(director.getNumberOfStates()*category.getNumberOfStates()); i++) {
					testDoubleElement = Math.random();
					testDouble[i] = testDoubleElement;
				}
				director.getTable().setData(testDouble);

				actor = new LabelledDCNode(this);
				actor.setName("actor");
				actor.setLabel("Actor");
				actor.setPosition(new Point(10,125));
				actor.setNumberOfStates(numberOfActorStates);
				for (int i=0; i<actor.getNumberOfStates(); i++) {
					actor.setStateLabel(i, ""+(i+1));
				}
				actor.addParent(category);
				testDouble = new double[actor.getNumberOfStates()*category.getNumberOfStates()];
				testDoubleElement = 0;
				for (int i=0; i<(actor.getNumberOfStates()*category.getNumberOfStates()); i++) {
					testDoubleElement = Math.random();
					testDouble[i] = testDoubleElement;
				}
				actor.getTable().setData(testDouble);
			}

			rating = new ContinuousChanceNode(this);
			rating.setName("rating");
			rating.setLabel("Rating");
			rating.setPosition(new Point(125,0));
			rating.addParent(community);
			rating.addParent(category);

			//set the means to randomized values:
			double random = 0;
			for (int i=0; i<(category.getNumberOfStates()*community.getNumberOfStates()); i++) {
				if (normalized) {
					random = (Math.random()*8)-4;
				}
				else {
					random = (Math.random()*4)+1;
				}
				rating.setAlpha(random, i);
			}

			//set all the variance values to 1:
			for (int i=0; i<(category.getNumberOfStates()*community.getNumberOfStates()); i++) {
				rating.setGamma(1.0, i);
			}

		}
		catch(ExceptionHugin eh) {
			eh.printStackTrace();
		}
	}

	/*
	*
	* Initializes the experience tables, and runs the predefined number of EM iterations and rounds
	*
	*/
	private void runTraining() {
		try {

			//set all prior experience entries to 1:
			user.getExperienceTable().setDataItem(0, 1);
			movie.getExperienceTable().setDataItem(0, 1);
			if (modelType.equals("zip")) {
				for (int i=0; i<zip.getExperienceTable().getSize(); i++) {
					zip.getExperienceTable().setDataItem(i, 1);
				}
			}
			if (modelType.equals("age")) {
				for (int i=0; i<age.getExperienceTable().getSize(); i++) {
					age.getExperienceTable().setDataItem(i, 1);
				}
			}
			if (modelType.equals("year")) {
				for (int i=0; i<year.getExperienceTable().getSize(); i++) {
					year.getExperienceTable().setDataItem(i, 1);
				}
			}
			if (modelType.equals("dir&act")) {
				for (int i=0; i<director.getExperienceTable().getSize(); i++) {
					director.getExperienceTable().setDataItem(i, 1);
				}
				for (int i=0; i<actor.getExperienceTable().getSize(); i++) {
					actor.getExperienceTable().setDataItem(i, 1);
				}
			}
			for (int i=0; i<community.getExperienceTable().getSize(); i++) {
				community.getExperienceTable().setDataItem(i, 1);
			}
			for (int i=0; i<category.getExperienceTable().getSize(); i++) {
				category.getExperienceTable().setDataItem(i, 1);
			}

			this.parseCases(trainingSet, this);
			this.compile();

			//run training on the non-continuous nodes
			trainUpperNetwork(trainingSet,iterations[0]);

			//run training on the continuous rating node
			for (int i=0; i<iterations[1]; i++) {
				trainRating();
			}

			//test the model against the provided test set:
			if (normalized) {
				countCorrectsNormalized();
			}
			else {
				countCorrectsUnNormalized();
			}

			//if more than one round of iterations are to be done, go here:
			for (int j=0; j<(iterations[2]-1); j++) {

				user.getExperienceTable().setDataItem(0, 1);
				movie.getExperienceTable().setDataItem(0, 1);
				if (modelType.equals("zip")) {
					for (int i=0; i<zip.getExperienceTable().getSize(); i++) {
						zip.getExperienceTable().setDataItem(i, 1);
					}
				}
				if (modelType.equals("age")) {
					for (int i=0; i<age.getExperienceTable().getSize(); i++) {
						age.getExperienceTable().setDataItem(i, 1);
					}
				}
				if (modelType.equals("year")) {
					for (int i=0; i<year.getExperienceTable().getSize(); i++) {
						year.getExperienceTable().setDataItem(i, 1);
					}
				}
				if (modelType.equals("dir&act")) {
					for (int i=0; i<director.getExperienceTable().getSize(); i++) {
						director.getExperienceTable().setDataItem(i, 1);
					}
					for (int i=0; i<actor.getExperienceTable().getSize(); i++) {
						actor.getExperienceTable().setDataItem(i, 1);
					}
				}
				for (int i=0; i<community.getExperienceTable().getSize(); i++) {
					community.getExperienceTable().setDataItem(i, 1);
				}
				for (int i=0; i<category.getExperienceTable().getSize(); i++) {
					category.getExperienceTable().setDataItem(i, 1);
				}

				trainUpperNetwork(trainingSet,iterations[0]);

				for (int i=0; i<iterations[1]; i++) {
					trainRating();
				}

				if (normalized) {
					countCorrectsNormalized();
				}
				else {
					countCorrectsUnNormalized();
				}
			}

		}
		catch(ExceptionHugin eh) {
			eh.printStackTrace();
		}
	}

	private void trainUpperNetwork(String trainingSet, int iterations) {
		try {
			
			if (iterations!=0) {
				this.setMaxNumberOfEMIterations(iterations);
			}
			this.learnTables();
		}
		catch(ExceptionHugin eh) {
			eh.printStackTrace();
		}
	}
	
	/**
	 * Does one iteration of EM learning on the continuous rating-node
	 *
	 */
	private void trainRating() {
		try {

			FileReader trainingReader = new FileReader(trainingSet);
			BufferedReader trainingBuffer = new BufferedReader(trainingReader);
			String temp = trainingBuffer.readLine();
			int userNr = 0;
			int zipNr = 0;
			int ageNr = 0;
			int movieNr = 0;
			int yearNr = 0;
			int directorNr = 0;
			int actorNr = 0;
			double ratingNr = 0;

			double[] ratingTable = new double[numberOfTrainingCases];

			double[][] comCatSum = new double[category.getNumberOfStates()][community.getNumberOfStates()];
			for (int i=0; i<comCatSum.length; i++) {
				for (int j=0; j<comCatSum[i].length; j++) {
					comCatSum[i][j] = 0;
				}
			}

			double[][] meanSum = new double[category.getNumberOfStates()][community.getNumberOfStates()];
			for (int i=0; i<meanSum.length; i++) {
				for (int j=0; j<meanSum[i].length; j++) {
					meanSum[i][j] = 0;
				}
			}

			double[][] varianceSum = new double[category.getNumberOfStates()][community.getNumberOfStates()];
			for (int i=0; i<varianceSum.length; i++) {
				for (int j=0; j<varianceSum[i].length; j++) {
					varianceSum[i][j] = 0;
				}
			}

			double[][][] comCatBeliefs = new double[numberOfTrainingCases][category.getNumberOfStates()][community.getNumberOfStates()];
			double[][] means = new double[category.getNumberOfStates()][community.getNumberOfStates()];

//			----------------------------STEP 1----------------------------------------
			int whileCounter = 0;
			temp = trainingBuffer.readLine();

			NodeList nodes = new NodeList();
			nodes.add(category);
			nodes.add(community);
			Table nodeTable;

			while (temp!=null) {
				//parsing the training set file:
				temp = temp.substring(1);
				userNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));

				if (modelType.equals("zip")) {
					temp = temp.substring(temp.indexOf(",")+2);
					zipNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));
				}

				if (modelType.equals("age")) {
					if (temp.substring(0,1).equals("\"")) {
						temp = temp.substring(temp.indexOf(",")+2);
						try {
							ageNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));
						}
						catch(NumberFormatException nfe) {
							ageNr = 0;
						}
					}
					else {
						temp = temp.substring(temp.indexOf(",")+1);
						ageNr = 0;
					}
				}

				temp = temp.substring(temp.indexOf(",")+2);
				movieNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));

				if (modelType.equals("year")) {
					if (temp.substring(0,1).equals("\"")) {
						temp = temp.substring(temp.indexOf(",")+2);
						try {
							yearNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));
						}
						catch(NumberFormatException nfe) {
							yearNr = 0;
						}
					}
					else {
						temp = temp.substring(temp.indexOf(",")+1);
						yearNr = 0;
					}
				}

				if (modelType.equals("dir&act")) {
					temp = temp.substring(temp.indexOf(",")+1);
					if (!temp.substring(0,1).equals(",")) {
						temp = temp.substring(1);
						try {
							directorNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));
						}
						catch(NumberFormatException nfe) {
							directorNr = 0;
						}
					}
					else {
						directorNr = 0;
					}
					temp = temp.substring(temp.indexOf(",")+1);
					if (!temp.substring(0,1).equals(",")) {
						temp = temp.substring(1);
						try {
							actorNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));
						}
						catch(NumberFormatException nfe) {
							actorNr = 0;
						}
					}
					else {
						actorNr = 0;
					}
				}

				temp = temp.substring(temp.indexOf(",")+1);
				ratingNr = Double.parseDouble(temp);

				ratingTable[whileCounter] = ratingNr;

				//registering the evidence into the network:
				movie.selectState(movieNr-1);
				user.selectState(userNr-1);
				if (modelType.equals("zip")) {
					zip.selectState(zipNr);
				}
				if (modelType.equals("age")) {
					if (ageNr!=0) {
						age.selectState(ageNr-1);
					}
				}
				if (modelType.equals("year")) {
					if (yearNr!=0) {
						year.selectState(yearNr-1);
					}
				}
				if (modelType.equals("dir&act")) {
					if (directorNr!=0) {
						director.selectState(directorNr-1);
					}
					if (actorNr!=0) {
						actor.selectState(actorNr-1);
					}
				}
				rating.enterValue(ratingNr);
				this.propagate(Domain.H_EQUILIBRIUM_SUM, Domain.H_EVIDENCE_MODE_NORMAL);
				//fetching the joint probabilities of community and category given the evidence:
				nodeTable = this.getMarginal(nodes);

				for (int i=0; i<community.getNumberOfStates(); i++) {
					for (int j=0; j<category.getNumberOfStates(); j++) {
						comCatBeliefs[whileCounter][j][i] = nodeTable.getDataItem(j*category.getNumberOfStates()+i);
					}
					initialize();
				}
				temp = trainingBuffer.readLine();
				whileCounter++;
			}

			//compute comCatSum:
			for (int k=0; k<comCatBeliefs.length; k++) {
				for (int i=0; i<comCatBeliefs[k].length; i++) {
					for (int j=0; j<comCatBeliefs[k][i].length; j++) {
						comCatSum[i][j] += comCatBeliefs[k][i][j];
					}
				}
			}

			//----------------------------STEP 2 - MEANS----------------------------------------

			for (int k=0; k<numberOfTrainingCases; k++) {
				for (int i=0; i<meanSum.length; i++) {
					for (int j=0; j<meanSum[i].length; j++) {
						meanSum[i][j]+= ratingTable[k]*comCatBeliefs[k][i][j];
					}
				}
			}

			int counter = 0;
			for (int i=0; i<meanSum.length; i++) {
				for (int j=0; j<meanSum[i].length; j++) {
					means[i][j] = meanSum[i][j]/(comCatSum[i][j]);
					rating.setAlpha(means[i][j],counter);
					counter++;
				}
			}

			//----------------------------STEP 3 - VARIANCES----------------------------------------

			for (int k=0; k<numberOfTrainingCases; k++) {
				for (int i=0; i<varianceSum.length; i++) {
					for (int j=0; j<varianceSum[i].length; j++) {
						varianceSum[i][j]+= (ratingTable[k]-means[i][j])*(ratingTable[k]-means[i][j])*comCatBeliefs[k][i][j];
					}
				}
			}

			counter = 0;
			for (int i=0; i<varianceSum.length; i++) {
				for (int j=0; j<varianceSum[i].length; j++) {
					rating.setGamma(varianceSum[i][j]/(comCatSum[i][j]),counter);
					counter++;
				}
			}

		}
		catch(ExceptionHugin eh) {
			eh.printStackTrace();
		}
		catch(IOException ioe) {
			ioe.printStackTrace();
		}
	}

	/**
	 * Uses the given test set to compute the MAE and RMSE of the network when the ratings are unnormalized
	 *
	 */
	private void countCorrectsUnNormalized() {
		try {

			FileReader testReader = new FileReader(testSet);
			BufferedReader testBuffer = new BufferedReader(testReader);
			String temp = testBuffer.readLine();
			int userNr = 0;
			int zipNr = 0;
			int ageNr = 0;
			int movieNr = 0;
			int yearNr = 0;
			int directorNr = 0;
			int actorNr = 0;
			int ratingNr = 0;
			double ratingMean = 0;
			double MAE = 0;
			double MSE = 0;
			double RMSE = 0;
			temp = testBuffer.readLine();
			while (temp!=null) {
				//parsing the test set file:
				temp = temp.substring(1);
				userNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));

				if (modelType.equals("zip")) {
					temp = temp.substring(temp.indexOf(",")+2);
					zipNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));
				}

				if (modelType.equals("age")) {
					try {
						if (temp.substring(0,1).equals("\"")) {
							temp = temp.substring(temp.indexOf(",")+2);
							ageNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));
						}
						else {
							ageNr = 0;
							temp = temp.substring(temp.indexOf(",")+1);
						}
					}
					catch(NumberFormatException nfe) {
						ageNr = 0;
					}
				}

				temp = temp.substring(temp.indexOf(",")+2);
				movieNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));

				if (modelType.equals("year")) {

					try {
						if (temp.substring(0,1).equals("\"")) {
							temp = temp.substring(temp.indexOf(",")+2);
							yearNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));
						}
						else {
							temp = temp.substring(temp.indexOf(",")+1);
							yearNr = 0;
						}
					}
					catch(NumberFormatException nfe) {
						yearNr = 0;
					}
				}

				if (modelType.equals("dir&act")) {
					temp = temp.substring(temp.indexOf(",")+1);
					if (!temp.substring(0,1).equals(",")) {
						temp = temp.substring(1);
						try {
							directorNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));
						}
						catch(NumberFormatException nfe) {
							directorNr = 0;
						}
					}
					else {
						directorNr = 0;
					}
					temp = temp.substring(temp.indexOf(",")+1);
					if (!temp.substring(0,1).equals(",")) {
						temp = temp.substring(1);
						try {
							actorNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));
						}
						catch(NumberFormatException nfe) {
							actorNr = 0;
						}
					}
					else {
						actorNr = 0;
					}
				}

				temp = temp.substring(temp.indexOf(",")+1);
				ratingNr = Integer.parseInt(temp);

				//register evidence:
				user.selectState(userNr-1);
				movie.selectState(movieNr-1);
				if (modelType.equals("zip")) {
					zip.selectState(zipNr);
				}
				if (modelType.equals("age")) {
					if (ageNr!=0) {
						age.selectState(ageNr-1);
					}
				}
				if (modelType.equals("year")) {
					if (yearNr!=0) {
						year.selectState(yearNr-1);
					}
				}
				if (modelType.equals("dir&act")) {
					if (directorNr!=0) {
						director.selectState(directorNr-1);
					}
					if (actorNr!=0) {
						actor.selectState(actorNr-1);
					}
				}
				this.propagate(Domain.H_EQUILIBRIUM_SUM, Domain.H_EVIDENCE_MODE_NORMAL);

				//read rating belief given the evidence
				ratingMean = rating.getMean();

				//measure MAE and MSE:
				MAE += Math.abs(ratingNr-ratingMean);
				MSE += (Math.abs(ratingNr-ratingMean))*(Math.abs(ratingNr-ratingMean));

//				initializing the network (retracting evidence):
				this.initialize();

				temp = testBuffer.readLine();
			}
			MAE = MAE/numberOfTestCases;
			MSE = MSE/numberOfTestCases;
			RMSE = Math.sqrt(MSE);
			results[0] = MAE;
			results[1] = RMSE;
		}
		catch(ExceptionHugin eh) {
			eh.printStackTrace();
		}
		catch(IOException ioe) {
			ioe.printStackTrace();
		}
	}

	/**
	 * 
	 * Uses the training set to compute the user-specific rating for each user 
	 *
	 */
	private void computeUserAverage() {
		try {
			//finding the file name of the file which contains the same data cases, but with unnormalized ratings:
			String[] tempTable = trainingSet.split("_Normalized");
			String tempTrainingSet = tempTable[0]+tempTable[1];

			FileReader averageReader = new FileReader(tempTrainingSet);
			BufferedReader averageBuffer = new BufferedReader(averageReader);
			String temp = averageBuffer.readLine();
			double[][] userAverage = new double[numberOfUsers][2];//1. er antall ratings og 2. er sum av ratings
			for (int i=0; i<userAverage.length; i++) {
				userAverage[i][0] = 0;
				userAverage[i][1] = 0;
			}
			int userNr = 0;
			int ratingNr = 0;
			temp = averageBuffer.readLine();
			while(temp!=null) {

				//parsing the training set file:
				temp = temp.substring(1);
				userNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));

				if (modelType.equals("zip")) {
					temp = temp.substring(temp.indexOf(",")+2);
				}
				else if (modelType.equals("age")) {
					if (temp.substring(0,1).equals("\"")) {
						temp = temp.substring(temp.indexOf(",")+2);
					}
					else {
						temp = temp.substring(temp.indexOf(",")+1);
					}
				}

				temp = temp.substring(temp.indexOf(",")+2);

				if (modelType.equals("year")) {
					if (temp.substring(0,1).equals("\"")) {
						temp = temp.substring(temp.indexOf(",")+2);
					}
					else {
						temp = temp.substring(temp.indexOf(",")+1);
					}
				}
				if (modelType.equals("dir&act")) {
					temp = temp.substring(temp.indexOf(",")+1);
					if (!temp.substring(0,1).equals(",")) {
						temp = temp.substring(1);
					}
					temp = temp.substring(temp.indexOf(",")+1);
					if (!temp.substring(0,1).equals(",")) {
						temp = temp.substring(1);
					}
				}

				temp = temp.substring(temp.indexOf(",")+1);
				ratingNr = Integer.parseInt(temp);

				userAverage[userNr-1][0]++;
				userAverage[userNr-1][1]+=ratingNr;

				temp = averageBuffer.readLine();
			}
			for (int i=0; i<averageTable.length; i++) {
				averageTable[i] = userAverage[i][1]/userAverage[i][0];
			}

		}
		catch(IOException ioe) {
			ioe.printStackTrace();
		}
	}
	
	/**
	 * Uses the given test set to compute the MAE and RMSE of the network when the ratings are normalized
	 *
	 */
	private void countCorrectsNormalized() {
		try {

			FileReader testReader = new FileReader(testSet);
			BufferedReader testBuffer = new BufferedReader(testReader);
			String temp = testBuffer.readLine();

			int userNr = 0;
			int zipNr = 0;
			int ageNr = 0;
			int movieNr = 0;
			int yearNr = 0;
			int directorNr = 0;
			int actorNr = 0;
			int ratingNr = 0;
			double ratingMean = 0;
			double MAE = 0;
			double MSE = 0;
			double RMSE = 0;
			temp = testBuffer.readLine();
			while (temp!=null) {
				//parsing the test set file:
				temp = temp.substring(1);
				userNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));

				if (modelType.equals("zip")) {
					temp = temp.substring(temp.indexOf(",")+2);
					zipNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));
				}

				if (modelType.equals("age")) {
					temp = temp.substring(temp.indexOf(",")+2);
					try {
						if (temp.substring(0,1).equals("\"")) {
							ageNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));
						}
						else {
							ageNr = 0;
						}
					}
					catch(NumberFormatException nfe) {
						ageNr = 0;
					}
				}

				temp = temp.substring(temp.indexOf(",")+2);
				movieNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));

				if (modelType.equals("year")) {
					temp = temp.substring(temp.indexOf(",")+2);
					try {
						if(temp.substring(0,1).equals("\"")) {
							yearNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));
						}
						else {
							yearNr = 0;
						}
					}
					catch(NumberFormatException nfe) {
						yearNr = 0;
					}
				}

				if (modelType.equals("dir&act")) {
					temp = temp.substring(temp.indexOf(",")+1);
					if (!temp.substring(0,1).equals(",")) {
						temp = temp.substring(1);
						try {
							directorNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));
						}
						catch(NumberFormatException nfe) {
							directorNr = 0;
						}
					}
					else {
						directorNr = 0;
					}
					temp = temp.substring(temp.indexOf(",")+1);
					if (!temp.substring(0,1).equals(",")) {
						temp = temp.substring(1);
						try {
							actorNr = Integer.parseInt(temp.substring(0,temp.indexOf("\"")));
						}
						catch(NumberFormatException nfe) {
							actorNr = 0;
						}
					}
					else {
						actorNr = 0;
					}
				}

				temp = temp.substring(temp.indexOf(",")+1);
				ratingNr = Integer.parseInt(temp);

//				registering evidence:
				user.selectState(userNr-1);
				movie.selectState(movieNr-1);
				if (modelType.equals("zip")) {
					zip.selectState(zipNr);
				}
				if (modelType.equals("age")) {
					if (ageNr!=0) {
						age.selectState(ageNr-1);
					}
				}
				if (modelType.equals("year")) {
					if (yearNr!=0) {
						year.selectState(yearNr-1);
					}
				}
				if (directorNr!=0) {
					director.selectState(directorNr-1);
				}
				if (actorNr!=0) {
					actor.selectState(actorNr-1);
				}
				this.propagate(Domain.H_EQUILIBRIUM_SUM, Domain.H_EVIDENCE_MODE_NORMAL);

				//get rating belief given the evidence:
				ratingMean = rating.getMean();

				//transform into the 5 scale rating:
				ratingMean += averageTable[userNr-1];

				//measure MAE og MSE:
				MAE += Math.abs(ratingNr-ratingMean);
				MSE += (Math.abs(ratingNr-ratingMean))*(Math.abs(ratingNr-ratingMean));

//				initializing the network (retracting evidence):
				this.initialize();

				temp = testBuffer.readLine();
			}
			MAE = MAE/numberOfTestCases;
			MSE = MSE/numberOfTestCases;
			RMSE = Math.sqrt(MSE);
			results[0] = MAE;
			results[1] = RMSE;
		}
		catch(ExceptionHugin eh) {
			eh.printStackTrace();
		}
		catch(IOException ioe) {
			ioe.printStackTrace();
		}
	}

	/**
	 * Method inherited from ParseListener, handles errors regarding the parsing of the data cases in Hugin
	 */
	public void parseError(int position, String msg) {
		System.out.println("Parsing the error: \""+msg+"\", which happened in position: "+position);
	}
}